from .unit import Unit
from .information_source import InformationSource
from .center import Center
from .persons import (
    Person,
    Patient,
    PatientForm,
    Examiner,
    ExaminerSerializer,
    PortalUserInfo,
    Profession,
)

from .examination import (
    Examination,
    ExaminationType,
    ExaminationTime,
    ExaminationTimeType,
)

from .data_file import *

from .patient_examination import PatientExamination

from .label import Label, LabelType, LabelSet

from .annotation import (
    ImageClassificationAnnotation,
    LegacyBinaryClassificationAnnotationTask,
    BinaryClassificationAnnotationTask,
)

from .legacy_data import (
    LegacyImage,
)

from .ai_model import (
    ModelMeta,
    ModelType,
    ActiveModel,
)

from .prediction import (
    ImageClassificationPrediction,
    LegacyVideoPredictionMeta,
    VideoPredictionMeta,
)

from .hardware import (
    EndoscopyProcessor,
    Endoscope,
    EndoscopeType,
)



from django.db import models

class EndoscopyProcessorManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class EndoscopyProcessor(models.Model):
    objects = EndoscopyProcessorManager()

    name = models.CharField(max_length=255)
    image_width = models.IntegerField()
    image_height = models.IntegerField()
    # image_fps = models.IntegerField()

    # Roi for endoscope image
    endoscope_image_x = models.IntegerField()
    endoscope_image_y = models.IntegerField()
    endoscope_image_width = models.IntegerField()
    endoscope_image_height = models.IntegerField()

    # Roi for examination date
    examination_date_x = models.IntegerField()
    examination_date_y = models.IntegerField()
    examination_date_width = models.IntegerField()
    examination_date_height = models.IntegerField()

    # Roi for examination time
    examination_time_x = models.IntegerField(blank=True, null=True)
    examination_time_y = models.IntegerField(blank=True, null=True)
    examination_time_width = models.IntegerField(blank=True, null=True)
    examination_time_height = models.IntegerField(blank=True, null=True)

    # Roi for patient first name
    patient_first_name_x = models.IntegerField()
    patient_first_name_y = models.IntegerField()
    patient_first_name_width = models.IntegerField()
    patient_first_name_height = models.IntegerField()

    # Roi for patient name
    patient_last_name_x = models.IntegerField()
    patient_last_name_y = models.IntegerField()
    patient_last_name_width = models.IntegerField()
    patient_last_name_height = models.IntegerField()

    # Roi for patient dob
    patient_dob_x = models.IntegerField()
    patient_dob_y = models.IntegerField()
    patient_dob_width = models.IntegerField()
    patient_dob_height = models.IntegerField()

    # Roi for endoscope type
    endoscope_type_x = models.IntegerField(blank=True, null=True)
    endoscope_type_y = models.IntegerField(blank=True, null=True)
    endoscope_type_width = models.IntegerField(blank=True, null=True)
    endoscope_type_height = models.IntegerField(blank=True, null=True)

    # Roi for endoscopy sn
    endoscope_sn_x = models.IntegerField(blank=True, null=True)
    endoscope_sn_y = models.IntegerField(blank=True, null=True)
    endoscope_sn_width = models.IntegerField(blank=True, null=True)
    endoscope_sn_height = models.IntegerField(blank=True, null=True)

    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    def get_roi_endoscope_image(self):
        return {
            "x": self.endoscope_image_x,
            "y": self.endoscope_image_y,
            "width": self.endoscope_image_width,
            "height": self.endoscope_image_height
        }
    
    def get_roi_examination_date(self):
        return {
            "x": self.examination_date_x,
            "y": self.examination_date_y,
            "width": self.examination_date_width,
            "height": self.examination_date_height
        }
    
    def get_roi_examination_time(self):
        return {
            "x": self.examination_time_x,
            "y": self.examination_time_y,
            "width": self.examination_time_width,
            "height": self.examination_time_height
        }
    
    def get_roi_patient_last_name(self):
        return {
            "x": self.patient_last_name_x,
            "y": self.patient_last_name_y,
            "width": self.patient_last_name_width,
            "height": self.patient_last_name_height
        }
    
    def get_roi_patient_first_name(self):
        return {
            "x": self.patient_first_name_x,
            "y": self.patient_first_name_y,
            "width": self.patient_first_name_width,
            "height": self.patient_first_name_height
        }
    
    def get_roi_patient_dob(self):
        return {
            "x": self.patient_dob_x,
            "y": self.patient_dob_y,
            "width": self.patient_dob_width,
            "height": self.patient_dob_height
        }
    
    def get_roi_endoscope_type(self):
        return {
            "x": self.endoscope_type_x,
            "y": self.endoscope_type_y,
            "width": self.endoscope_type_width,
            "height": self.endoscope_type_height
        }
    
    def get_roi_endoscopy_sn(self):
        return {
            "x": self.endoscope_sn_x,
            "y": self.endoscope_sn_y,
            "width": self.endoscope_sn_width,
            "height": self.endoscope_sn_height
        }
    
    def get_rois(self):
        return {
            "endoscope_image": self.get_roi_endoscope_image(),
            "examination_date": self.get_roi_examination_date(),
            "examination_time": self.get_roi_examination_time(),
            "patient_first_name": self.get_roi_patient_first_name(),
            "patient_last_name": self.get_roi_patient_last_name(),
            "patient_dob": self.get_roi_patient_dob(),
            "endoscope_type": self.get_roi_endoscope_type(),
            "endoscope_sn": self.get_roi_endoscopy_sn()
        }



from django.db import models
from ..person import Person
from rest_framework import serializers

class Examiner(Person):
    center = models.ForeignKey('Center', on_delete=models.CASCADE, blank=True, null=True)

    def __str__(self):
        return self.first_name + " " + self.last_name
    

class ExaminerSerializer(serializers.ModelSerializer):
    
    class Meta:
        model = Examiner
        fields = '__all__'


from abc import abstractmethod
from django.db import models

class Person(models.Model):
    """
    Abstract base class for a person.

    Attributes:
        first_name (str): The first name of the person.
        last_name (str): The last name of the person.
        dob (date): The date of birth of the person.
        gender (str): The gender of the person.
        email (str): The email address of the person.
        phone (str): The phone number of the person.
    """

    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    dob = models.DateField("Date of Birth", blank=True, null=True)
    gender = models.CharField(
        max_length=10, choices=[('Male', 'Male'), ('Female', 'Female'), ('Other', 'Other')],
        blank=True, null=True
    )
    email = models.EmailField(max_length=255, blank=True, null=True)
    phone = models.CharField(max_length=255, blank=True, null=True)
    is_real_person = models.BooleanField(default=True)

    @abstractmethod
    def __str__(self):
        pass

    class Meta:
        abstract = True




from .model_meta import ModelMeta

# get latest model meta by model name

# TODO MOVE THIS TO A QUERY FILE
def get_latest_model_meta_by_model_name(model_name):
    model_meta = ModelMeta.objects.filter(name=model_name).order_by('-version').first()
    return model_meta



from django.db import models

class ExaminationTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class ExaminationType(models.Model):
    objects = ExaminationTypeManager()
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return self.name_en
    
    def natural_key(self):
        return (self.name,)
    



from agl_base_db.models import (
    ModelMeta,
)

def get_latest_model_by_name(model_name):
    """
    Expects model name. Fetches models by name from database, sorts by version and returns latest version.
    """
    models = ModelMeta.objects.filter(name=model_name).order_by('-version')
    if len(models) == 0:
        return None
    else:
        return models[0]



from .ai_model import *
from .annotation import *
from .center import *
from .examination import *
from .frame import *
from .hardware import *
from .label import *
from .patient import *
from .report_file import *
from .video import *


from rest_framework import serializers
from agl_base_db.models import Video, LegacyVideo, VideoImportMeta, LegacyLabelVideoSegment, LabelVideoSegment

class VideoSerializer(serializers.ModelSerializer):
    class Meta:
        model = Video
        fields = '__all__'

class LegacyVideoSerializer(serializers.ModelSerializer):
    class Meta:
        model = LegacyVideo
        fields = '__all__'

class VideoImportMetaSerializer(serializers.ModelSerializer):
    class Meta:
        model = VideoImportMeta
        fields = '__all__'

class LabelVideoSegmentSerializer(serializers.ModelSerializer):
    class Meta:
        model = LabelVideoSegment
        fields = '__all__'

class LegacyLabelVideoSegmentSerializer(serializers.ModelSerializer):
    class Meta:
        model = LegacyLabelVideoSegment
        fields = '__all__'



from rest_framework import serializers
from agl_base_db.models import (
    Examination,
    ExaminationTimeType,
    ExaminationTime,
    ExaminationType,
    PatientExamination
)

class ExaminationTimeTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExaminationTimeType
        fields = '__all__'

class ExaminationTimeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExaminationTime
        fields = '__all__'

class ExaminationTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExaminationType
        fields = '__all__'

class ExaminationSerializer(serializers.ModelSerializer): 
    class Meta:
        model = Examination
        fields = '__all__'

class PatientExaminationSerializer(serializers.ModelSerializer):
    class Meta:
        model = PatientExamination
        fields = '__all__'


from agl_base_db.models import Video, PatientExamination
from typing import Optional

def get_videos() -> Video:
    """
    Returns all Video objects from the database.
    """
    return Video.objects.all()

def get_video_by_id(id) -> Optional[Video]:
    """Retrieve a Video object by its id.

    Args:
        id (int): The id of the video to retrieve.

    Returns:
        Optional[Video]: The Video object with the given id, or None if it does not exist.
    """
    return Video.objects.get(id=id)

def get_video_without_examination() -> Video:
    """
    Returns all Video objects from the database without an examination.
    """
    return Video.objects.filter(patient_examination__isnull=True)

def get_video_without_patient() -> Video:
    """
    Returns all Video objects from the database without a patient.
    """
    return Video.objects.filter(patient__isnull=True)



from rest_framework import serializers
from agl_base_db.models import (
    Endoscope,
    EndoscopeType,
    EndoscopyProcessor,
)

class EndoscopeSerializer(serializers.ModelSerializer):
    class Meta:
        model = Endoscope
        fields = '__all__'

class EndoscopyProcessorSerializer(serializers.ModelSerializer):
    class Meta:
        model = EndoscopyProcessor
        fields = '__all__'

class EndoscopeTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = EndoscopeType
        fields = '__all__'


from ..base_classes import AbstractVideo
from django.db import models

from agl_base_db.models.data_file.frame import Frame
from agl_base_db.models.data_file.frame import LegacyFrame

BATCH_SIZE = 1000

class Video(AbstractVideo):
    import_meta = models.OneToOneField('VideoImportMeta', on_delete=models.CASCADE, blank=True, null=True)
    def get_video_model(self):
        return Video

    def get_frame_model(self):
        return Frame


class LegacyVideo(AbstractVideo):
    file = models.FileField(upload_to="legacy_videos", blank=True, null=True)

    def get_video_model(self):
        return LegacyVideo

    def get_frame_model(self):
        return LegacyFrame



from django.db import models
from ..center import Center
import hashlib
from datetime import date, time

class ReportFile(models.Model):
    pdf = models.FileField(upload_to="raw_report_pdfs", blank=True, null=True)
    pdf_hash = models.CharField(max_length=255, unique=True)
    center = models.ForeignKey(Center, on_delete=models.CASCADE)
    meta = models.JSONField(blank=True, null=True)
    text = models.TextField(blank=True, null=True)
    text_anonymized = models.TextField(blank=True, null=True)
    patient = models.ForeignKey("Patient", on_delete=models.CASCADE, blank=True, null=True)
    examiner = models.ForeignKey("Examiner", on_delete=models.CASCADE, blank=True, null=True)
    date = models.DateField(blank=True, null=True)
    time = models.TimeField(blank=True, null=True)

    def get_pdf_hash(self):
        pdf = self.pdf
        pdf_hash = None

        if pdf:
            # Open the file in binary mode and read its contents
            with pdf.open(mode='rb') as f:
                pdf_contents = f.read()
                # Create a hash object using SHA-256 algorithm
                hash_object = hashlib.sha256(pdf_contents, usedforsecurity=False)
                # Get the hexadecimal representation of the hash
                pdf_hash = hash_object.hexdigest()
                assert len(pdf_hash) <= 255, "Hash length exceeds 255 characters"

        return pdf_hash

    def initialize_metadata_in_db(self, report_meta=None):
        if not report_meta:
            report_meta = self.meta
        self.set_examination_date_and_time(report_meta)
        self.patient, created = self.get_or_create_patient(report_meta)
        self.examiner, created = self.get_or_create_examiner(report_meta)
        self.save()    

    def get_or_create_patient(self, report_meta=None):
        from ..persons import Patient
        if not report_meta:
            report_meta = self.meta
        patient_first_name = report_meta['patient_first_name']
        patient_last_name = report_meta['patient_last_name']
        patient_dob = report_meta['patient_dob']

        patient, created = Patient.objects.get_or_create(
            first_name=patient_first_name,
            last_name=patient_last_name,
            dob=patient_dob
        )

        return patient, created
    
    def get_or_create_examiner(self, report_meta= None):
        from ..persons import Examiner
        if not report_meta:
            report_meta = self.meta
        examiner_first_name = report_meta['examiner_first_name']
        examiner_last_name = report_meta['examiner_last_name']
        examiner_center = self.center

        examiner, created = Examiner.objects.get_or_create(
            first_name=examiner_first_name,
            last_name=examiner_last_name,
            center=examiner_center
        )

        return examiner, created
        
    def set_examination_date_and_time(self, report_meta=None):
        if not report_meta:
            report_meta = self.meta
        examination_date_str = report_meta['examination_date']
        examination_time_str = report_meta['examination_time']

        if examination_date_str:
            # TODO: get django DateField compatible date from string (e.g. "2021-01-01")
            self.date = date.fromisoformat(examination_date_str)
        if examination_time_str:
            # TODO: get django TimeField compatible time from string (e.g. "12:00")
            self.time = time.fromisoformat(examination_time_str)
        





from rest_framework import serializers
from agl_base_db.models import (
    Patient,
)

class PatientSerializer(serializers.ModelSerializer):
    class Meta:
        model = Patient
        fields = '__all__'




# django command to register a new AI model
# expects path to model_meta.json file
# example model_meta: {
# "name": "multilabel_classification",
# "version": 0,
# "model_type": "multilabel_classification", # name of modeltype, is unique
# "labelset": "multilabel_classification", #labelset name, combination of name and version is unique
# "labelset_version": 0,
# "weights_path": "weights/multilabel_classification_0.pth", # path to weights file
#}

from django.core.management.base import BaseCommand
from django.core.files import File
from agl_base_db.models import ModelMeta, ModelType, LabelSet
import json
from pathlib import Path
from icecream import ic

class Command(BaseCommand):
    """
    Registers a new AI model in the database.
    """
    help = 'Registers a new AI model in the database.'

    def add_arguments(self, parser):
        parser.add_argument('model_meta_path', type=str)

    def handle(self, *args, **options):
        model_meta_path = Path(options['model_meta_path']).resolve()

        with open(model_meta_path, 'r') as f:
            model_meta = json.load(f)

        # get or create model type
        model_type = ModelType.objects.get(name=model_meta['model_type'])

        # get or create labelset
        labelset = LabelSet.objects.get(name=model_meta['labelset'], version=model_meta['labelset_version'])

        # Handle weights file
        weights_path = model_meta['weights_path']
        # weights path is realative to model_meta_path
        weights_path = model_meta_path.parent / weights_path

        assert weights_path.exists(), f"weights file at {weights_path} does not exist"

        # Make sure the path is correct and the file exists
        try:
            with open(weights_path, 'rb') as file:
                model_name_string = f"{model_meta['name']}_{model_meta['version']}"
                weights = File(file, name = model_name_string)
                # Create ModelMeta instance
                model_meta_instance = ModelMeta.objects.create(
                    name=model_meta['name'],
                    version=model_meta['version'],
                    type=model_type,
                    labelset=labelset,
                    weights=weights,
                    description=model_meta.get('description', '')  # Assuming description is optional
                )
                print(f"Successfully registered model {model_meta_instance}")
        except IOError:
            print(f"Failed to open weights file at {weights_path}. Make sure the file exists.")





from rest_framework import serializers
from agl_base_db.models import ReportFile

class ReportFileSerializer(serializers.ModelSerializer):
    class Meta:
        model = ReportFile
        fields = '__all__'



from django.db import models

class VideoImportMeta(models.Model):
    processor = models.ForeignKey('EndoscopyProcessor', on_delete=models.CASCADE)
    endoscope = models.ForeignKey('Endoscope', on_delete=models.CASCADE, blank=True, null=True)
    center = models.ForeignKey('Center', on_delete=models.CASCADE)
    video_anonymized = models.BooleanField(default=False)
    video_patient_data_detected = models.BooleanField(default=False)
    outside_detected = models.BooleanField(default=False)
    patient_data_removed = models.BooleanField(default=False)
    outside_removed = models.BooleanField(default=False)

    def __str__(self):
        result_html = ""

        result_html += f"Processor: {self.processor.name}<br>"
        result_html += f"Endoscope: {self.endoscope.name}<br>"
        result_html += f"Center: {self.center.name}<br>"
        result_html += f"Video anonymized: {self.video_anonymized}<br>"
        result_html += f"Video patient data detected: {self.video_patient_data_detected}<br>"
        result_html += f"Outside detected: {self.outside_detected}<br>"
        result_html += f"Patient data removed: {self.patient_data_removed}<br>"
        result_html += f"Outside removed: {self.outside_removed}<br>"
        return result_html
    



from agl_base_db.models import Patient

def get_patients() -> Patient:
    """
    Returns all Patient objects from the database.
    """
    return Patient.objects.all()

def get_patients_without_dob() -> Patient:
    """
    Returns all Patient objects from the database without a date of birth.
    """
    return Patient.objects.filter(dob__isnull=True)




import hashlib

def get_video_hash(video_path):
    """
    Get the hash of a video file.
    """
    # Open the video file in read-binary mode:
    with open(video_path, 'rb') as f:
        # Create the hash object, passing in the video contents for hashing:
        hash_object = hashlib.sha256(f.read())
        # Get the hexadecimal representation of the hash
        video_hash = hash_object.hexdigest()
        assert len(video_hash) <= 255, "Hash length exceeds 255 characters"

    return video_hash

